// Copyright (c) 2010-2026, Lawrence Livermore National Security, LLC. Produced
// at the Lawrence Livermore National Laboratory. All Rights reserved. See files
// LICENSE and NOTICE for details. LLNL-CODE-443271.
//
// This file is part of the GLVis visualization tool and library. For more
// information and source code availability see https://glvis.org.
//
// GLVis is free software; you can redistribute it and/or modify it under the
// terms of the BSD-3 license. We welcome feedback and contributions, see file
// CONTRIBUTING.md for details.

#ifndef GLVIS_VISUAL_HPP
#define GLVIS_VISUAL_HPP

// Visualization header file

#include "vsdata.hpp"
#include "material.hpp"
#include "aux_vis.hpp"
#include "openglvis.hpp"
#include "vssolution.hpp"
#include "vssolution3d.hpp"
#include "vsvector.hpp"
#include "vsvector3d.hpp"
#include "threads.hpp"
#include "gl/types.hpp"

#endif // GLVIS_VISUAL_HPP
